/**************************************************************

cmunipack.h (C-Munipack project)
Public header file
Copyright (C) 2003 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#ifndef CMUNIPACK_H
#define CMUNIPACK_H

#include <fitsio.h>

/******************* Public constats ************************/

// Image format identifiers
#define MPK_FORMAT_UNKNOWN    0					// Unknown format
#define MPK_FORMAT_FITS       1					// FITS format
#define MPK_FORMAT_SBIG       2					// SBIG format
#define MPK_FORMAT_OES        3					// OES Astro format

// Output format identifiers
#define MPK_FORMAT_DIFFMAG    0					// Differential magnitudes (default)
#define MPK_FORMAT_INSTMAG    1					// Instrumental magnitudes
#define MPK_FORMAT_READALL    2					// Readall format
#define MPK_FORMAT_TRACKLST   3					// Track list
#define MPK_FORMAT_STDDEV     4					// Standard deviations

// Output format modifiers
#define MPK_FORMAT_JDHEL      0x0100			// Include heliocentric julian dates
#define MPK_FORMAT_HELCOR     0x0200			// Include heliocentric correction
#define MPK_FORMAT_AIRMASS    0x0400			// Include airmass coefficient

// Output level
#define MPK_LEVEL_NORMAL	  0					// Normal level (default)
#define MPK_LEVEL_QUIET      -1					// Suppress messages except error reports
#define MPK_LEVEL_DEBUG       1					// Verbose output, for debugging purposes

// Conversion modes
#define MPK_CONVERT_NORMAL	  0					// Normal conversion
#define MPK_CONVERT_HEADER    1					// Create only header of FITS file 
#define MPK_CONVERT_INFO      2					// Print short info only 
#define MPK_CONVERT_DETAILS   3					// Print detailed info only 

// Dark-frame correction modes
#define MPK_DARK_STANDARD     0					// Standard calibration scheme
#define MPK_DARK_SCALING      1					// Use dark-frame scaling

// Master-dark modes
#define MPK_MDARK_MASTERDARK  0x0000			// Make master dark frame	
#define MPK_MDARK_MASTERBIAS  0x0001			// Make master bias frame
#define MPK_MDARK_SCALABLE    0x0010			// Make scalable dark frame

/********************* Error codes **************************/
// Fitsio library uses error codes less than 1000           

// Common error codes
#define MPK_MEMORY            1001				// Insufficient memory
#define MPK_CANNOT_OPEN_SRC   1003				// Cannot open the source file
#define MPK_CANNOT_OPEN_OUT   1004				// Cannot open the destination file
#define MPK_CANNOT_OPEN_PAR   1005				// Cannot open the parameter file
#define MPK_NO_DIRFILE        1006				// Dirfile not found
#define MPK_NO_INPUT_FILES    1007				// No input files
#define MPK_NO_DATA           1008				// No data loaded in memory
#define MPK_UNKNOWN_FORMAT    1009				// Unknown format of source file
#define MPK_INVALID_SIZE      1010				// Invalid dimensions of image
#define MPK_INVALID_PAR       1011				// Invalid value of parameter
#define MPK_INVALID_DATA      1012				// Invalid data in source file
#define MPK_INVALID_HEADER    1013				// Error in header of the source file
#define MPK_INVALID_DATE      1014				// Invalid date/time format
#define MPK_ERROR             1015				// General error
#define MPK_KEY_NOT_FOUND     1016				// Key not found
#define MPK_DATA_OVERFLOW     1017				// An overflow has been occurred during computation
#define MPK_DATA_UNDERFLOW    1018				// An underflow has been occurred during computation
#define MPK_CLOSED_FILE       1019				// Operation not allowed on closed file
#define MPK_FRAME_NOT_FOUND   1020				// Frame not found.
#define MPK_READ_ERROR        1023				// Error while reading file
#define MPK_READ_ONLY         1024				// File is open in read-only mode
#define MPK_INVALID_OBJRA     1025				// Invalid R.A. format
#define MPK_INVALID_OBJDEC    1026				// Invalid DEC. format
#define MPK_INVALID_OBSLON    1027				// Invalid longitude format
#define MPK_INVALID_OBSLAT    1028				// Invalid latitude format
#define MPK_MISSING_OBJCOORDS 1029				// Missing object's coordinates
#define MPK_MISSING_OBSCOORDS 1030              // Missing observer's coordinates
#define MPK_INVALID_EXPTIME   1031				// Invalid exposure duration

// Darkbat utility
#define DRK_CANNOT_OPEN_DRK   1101				// Dark frame not found
#define DRK_DIFFERENET_SIZE   1102				// Dimensions of dark-frame and scientific image are different
#define DRK_INVALID_EXPTIME   1103				// Invalid exposure duration

// Flatbat utility
#define FLT_CANNOT_OPEN_FLT   1201				// Flat frame not found
#define FLT_DIFFERENET_SIZE   1202				// Dimensions of flat-frame and scientific image are different
#define FLT_MEAN_ZERO         1203				// Mean value of flat frame is zero (can't divide by zero)
#define FLT_DIVZERO           1204				// Zero value has been occurred on flat frame (can't divide by zero)

// Munimat utility
#define MAT_CANNOT_OPEN_REF   1301				// Reference file not found
#define MAT_FEW_POINTS_REF    1302				// Too few stars in the reference file
#define MAT_FEW_POINTS_SRC    1303				// Too few stars in the source file
#define MAT_MATCH_NOT_FOUND   1304				// Coincidences not found
#define MAT_XML_PARSE         1305				// Syntax error in the catalogue file

// Autoflat utility
#define AFL_DIFFERENET_SIZE   1401				// Input frames are not compatible (different sizes)

// Meandark utility
#define MDK_DIFFERENET_SIZE   1501				// Input frames are not compatible (different sizes)
#define MDK_CANNOT_OPEN_BIAS  1502				// Bias frame not found
#define MDK_DIFF_BIASSIZE     1503              // Bias frame is not compatible (different sizes)

// Varfind utility
#define MFI_TOO_FEW_DATA      1601				// Too few data in source files
#define MFI_REF_NOT_FOUND     1602				// Refererence star was not found

// Kombine utility
#define KMB_MAT_NOT_FOUND     1701              // Matched photometry file not found
#define KMB_DIFFERENET_SIZE   1702				// Input frames are not compatible (different sizes)
#define KMB_OUT_NOT_OPEN      1703              // Output file is not open

// Munilist utility
#define MLS_SYNTAX_ERROR	  1801				// String syntax error

// Biasbat utility
#define MPK_CANNOT_OPEN_BIAS  1901				// Bias frame not found
#define MPK_DIFF_BIAS_SIZE    1902				// Dimensions of bias frame and scientific image are different

/******************* Public data types ***********************/

// CallBack function type. Parameter points to the memory buffer with
// nullterminated message.
typedef void cmpack_cbtype(const char *);

#ifdef __cplusplus
extern "C" {
#endif

/******************** Common functions ***********************/

// Prints CMunipack's version to the specified buffer
//	buf - pointer to the buffer
//  buflen - length of the buffer in bytes
// Returns nothing.
void cmpack_getversionid(char *buf, int buflen);
    
// Translates error code to its human readable representation.
//	buf - pointer to the buffer
//  buflen - length of the buffer in bytes
//	code - error code
// Returns nothing.
void cmpack_formaterror(char *buf, int buflen, int code);

// Sets output level.
//	cbproc - callback function
//  level - see MPK_LEVEL constants
// Returns previous output level
int cmpack_setoutput(cmpack_cbtype *cbproc, int level);

/******************** Conversion of coordinates **********************/

// Converts the string to R.A.
//	buf - input string
//  ra  - pointer to variable which receives the right ascension in hours
// Returns zero on success or MPK_INVALID_PAR error code on failure.
int  cmpack_strtora(const char *buf, double *ra);

// Converts the string to DEC.
//	buf - input string
//  dec - pointer to variable which receives the declination in degrees
// Returns zero on success or MPK_INVALID_PAR error code on failure.
int  cmpack_strtodec(const char *buf, double *dec);

// Converts the R.A. value to the string (HH MM SS)
//  ra  - right ascension value in hours
//	buf - pointer to memory buffer (at least 9 bytes long)
// Returns nothing.
void cmpack_ratostr(double ra, char *buf);

// Converts the DEC. value to the string (+DDD MM SS)
//  dec - declination value in degrees
//	buf - pointer to memory buffer (at least 11 bytes long)
// Returns nothing.
void cmpack_dectostr(double dec, char *buf);

// Converts string to latitude.
//	buf - input string
//  lat - pointer to variable which receives the latitude in degrees
//			(positive to the north, negative to the south)
// Returns zero on success or MPK_INVALID_PAR error code on failure.
int  cmpack_strtolat(const char *buf, double *lat);
	 
// Converts string to longitude.
//	buf - input string
//  lat - pointer to variable which receives the longitude in degrees
//        (positive to the east, negative to the west)
// Returns zero on success or MPK_INVALID_PAR error code on failure.
int  cmpack_strtolon(const char *buf, double *lon);

// Converts the latitude to the string (N DDD MM SS)
//  lat - latitude in degrees
//	buf - pointer to memory buffer (at least 12 bytes long)
// Returns nothing.
void cmpack_lattostr(double lat, char *buf);

// Converts the longitude to the string (E DDD MM SS)
//  lon - longitude in degrees
//	buf - pointer to memory buffer (at least 12 bytes long)
// Returns nothing.
void cmpack_lontostr(double lon, char *buf);

// Converts the azimuth to the string (DDD MM SS (XXX))
//  az - azimuth in degrees
//	buf - pointer to memory buffer (at least 16 bytes long)
// Returns nothing.
void cmpack_aztostr(double az, char *buf);

// Converts the azimuth to the string (DDD MM SS (XXX))
//  az - azimuth in degrees
//	buf - pointer to memory buffer (at least 16 bytes long)
// Returns nothing.
void cmpack_alttostr(double az, char *buf);

/******************** Date & time conversion functions ********************/

// Parses string and returns the parts of the date.
//	datestr - date string (YYYY-MM-DD)
//	year, month, day - receives date parts
// Returns nothing.
int cmpack_strtodate(const char *datestr, int *year, int *month, int *day);

// Parses string and returns the parts of the time.
//	timestr - date string (HH:MM:SS.SSS)
//	hour, minute, second, milisecond - receives time parts
// Returns nothing.
int cmpack_strtotime(const char *timestr, int *hour, int *minute, int *second, int *milisecond);

// Converts date and time strings to julian date
//	datestr - date string (YYYY-MM-DD)
//	timestr - date string (HH:MM:SS.SSS)
// Returns julian date or zero on failure.
double cmpack_strtojd(const char *datestr, const char *timestr);

// Converts date and time parts to julian date
//	year, month, day - date parts
//	hour, minute, second, milisecond - time parts
// Returns julian date or zero on failure.
double cmpack_encodejd(int year, int month, int day, int hour, int minute, int second, int milisecond);

// Prints date parts to the string
//	year, month, day - date parts
//  datestr - pointer to the buffer which receives the date string (at least 11 bytes)
// Returns nothing.
void cmpack_datetostr(int year, int month, int day, char *datestr);

// Prints time parts to the string
//	hour, minute, second, milisecond - time parts
//  timestr - pointer to the buffer which receives the time string (at least 13 bytes)
// Returns nothing.
void cmpack_timetostr(int hour, int minute, int second, int msec, char *timestr);

// Converts julian date to date and time parts
//	jd - julian date
//	year, month, day - receives date parts
//	hour, minute, second, milisecond - receives time parts
// Returns nothing.
void cmpack_decodejd(double jd, int *year, int *month, int *day, int *hour, int *minute, int *second, int *milisecond);
	
// Converts julian date to date and time strings
//	jd - julian date
//  datestr - pointer to the buffer which receives the date string (at least 11 bytes)
//  timestr - pointer to the buffer which receives the time string (at least 13 bytes)
// Returns nothing.
void cmpack_jdtostr(double jd, char *datestr, char *timestr);

/************************ Mathematical functions ***********************************/

// Computes robust mean and standard deviation of 'n' elements in 'A' array. 
//	n - number of elements
//	A - array of elements
//	mean - receives robust mean
//	sig - receives standard deviation
// Returns nothing.
void cmpack_robustmean(int n, float *A, float *mean, float *sig);
void cmpack_robustmeanD(int n, double *A, double *mean, double *sig);

// Computes robust mean and standard deviation of ccd frame (much faster but less precise method)
//	nx, ny - number of rows, columns
//	ccd - pointer to image data
//  nmax - maximal number of pixels used in computation
//  lobad, hibad - minimal and maximal valid values of pixels in ADU
//  skymed - receives sky background value (robust mean of pixel values)
//  skysig - receives sky noise value (standard deviation of pixel values)
// Returns nothing.
void cmpack_fastsky(int nx, int ny, float *ccd, int nmax, float lobad, float hibad, float *skymed, float *skysig);
void cmpack_fastskyD(int nx, int ny, double *ccd, int nmax, double lobad, double hibad, double *skymed, double *skysig);

// Computes robust mean and standard deviation of ccd frame (slower but more precise method)
//	nx, ny - number of rows, columns
//	ccd - pointer to image data
//  skymed - receives sky background value (robust mean of pixel values)
//  skysig - receives sky noise value (standard deviation of pixel values)
// Returns nothing.
void cmpack_fullsky(int nx, int ny, float *ccd, float *skymed, float *skysig);
void cmpack_fullskyD(int nx, int ny, double *ccd, double *skymed, double *skysig);

// Computes Greenwich mean sidereal time.
//	jd - julian date
// Returns Greenwich mean sidereal time in days.
double cmpack_siderealtime(double jd);

// Computes Airmass coefficient.
//	jd - julian date
//	obj_ra, obj_dec - equatorial coordinates of the object (hours, degrees)
//	obs_lon, obs_lat - geographical coordinates of the observer (degrees)
// Returns airmass coefficient (X=1 for object in zenith)
double cmpack_airmass(double jd, double obj_ra, double obj_dec, double obs_lon, double obs_lat);

// Translates equatorial coordinates to the ecliptical coordinates
//	obj_ra, obj_dec - equatorial coordinates (hours, degrees)
//	obj_la, obj_be - ecliptical coordinates (degrees)
// Returns nothing.
void cmpack_rdtolb(double obj_ra, double obj_de, double *obj_la, double *obj_be);

// Computes position of the Sun.
//	jd - julian date
//	sun_ls, sun_rs - ecliptical longitude and distance of the Sun (degrees, AU)
// Returns nothing.
void cmpack_sun(double jd, double *sun_ls, double *sun_rs);

// Computes heliocentric correction
//	obj_la, obj_be - ecliptical coordinates of the object (degrees)
//	sun_ls, sun_rs - ecliptical longitude and distance of the Sun (degrees, AU)
// Returns heliocentric correction in days.
double cmpack_hcor(double obj_la, double obj_be, double sun_ls, double sun_rs);

// Transforms equatorial coordinates to horizontal system 
// jd - julian date (UT)
//	ra, dec  - equatorial coordinates of object (hours, degrees)
//	lon, lat - geographical coordinates of the observer (degrees)
//	az, alt  - horizontal coordinates (degrees)
// Returns nothing
void cmpack_azalt(double jd, double obj_ra, double obj_dec, double obs_lon, double obs_lat, double *az, double *alt);

// Returns path to the directory with default configuration files
const char *cmpack_shareddir(void);

/* AUTOFLAT module */

int aflat_init(const char *optfile);
    // Initialize variables and set parameters
int aflat_read(const char *infile);
    // Read input image
int aflat_write(const char *outfile, double *skymed, double *skysig);
    // Compute output image and write it to the file
int aflat_clean(void);
    // Free allocated memory

/* BIASBAT module */

int bias_init(const char *bias);
    // Initialize variables and load the bias-frame
int bias(const char *sci, const char *out);
    // Run bias correction of 'sci' file into 'out' file
int bias_clean(void);
    // Free allocated memory

/* DARKBAT module */

int dark_init(const char *drk, int mode);
    // Initialize variables and load the dark-frame
int dark(const char *sci, const char *out);
    // Run dark correction of 'sci' file into 'out' file
int dark_clean(void);
    // Free allocated memory

/* FLATBAT module */

int flat_init(const char *flt);
    // Initialize variables and load the flat-frame
int flat(const char *sci, const char *out);
    // Run flat correction of 'sci' file into 'out' file
int flat_clean(void);
    // Free allocated memory

/* KONVE module */

int konv_init(const char *parfile, int flipx, int flipy, double tcor, int mode);
    // Initializes the variables and reads the parameters
int konv_format(const char *filename, int *format);
    // Determines the format of the file
int konv(const char *filename, const char *out);
    // Converts the file to FITS format
int konv_clean(void);
    // Frees allocated memory

int konv_open(void **handle, const char *filename);
	// Open an image file (autodetect type)
int konv_openf(void **handle, const char *filename, int format);
	// Opens an image file of known type
int konv_gets(void *handle, const char *key, char *buf, int buflen);
	// Get value as string
int konv_getl(void *handle, const char *key, long *val);
	// Get value as integer number
int konv_getd(void *handle, const char *key, double *val);
	// Get value as floating point number
int konv_gkyn(void *handle, int index, char *key, int keylen, char *val, int vallen, char *com, int comlen);
	// Get (key, value, comment) triplet
int konv_g2dd(void *handle, double *buf, int flipx, int flipy);
	// Get image data
int konv_close(void *handle);
	// Close file

/* MUNIMATCH module */

int match_init(const char *reffile, const char *optfile);
    // Initializes the variables and reads the reference and parameter file
int match(const char *infile, const char *outfile, int *mstar);
    // Reads input file, finds crossreferences and writes the output file
int match_clean(void);
    // Frees allocated memory

/* MEANDARK module */

int mdark_init(const char *optfile, int flags, const char *biasfile);
    // Initialization of variables
int mdark_read(const char *infile);
    // Read input file
int mdark_write(const char *outfile);
    // Compute 'master-dark' frame and write it to the output file
int mdark_clean(void);
    // Free allocated memory

/* MUNIFIND module */

int mfind_init(const char *optfile, int ap, int thr);
    // Initialize parameteres and open output file
int mfind_readsrt(const char *srtfile);
    // Read data from photometry file
int mfind_readall(const char *srcfile);
    // Read data from all-stars file
int mfind_compute(int ref);
    // Proved vypocet
int mfind_getpar(int *nstars, int *nimages, int *ref);
    // Returns parameters
int mfind_get(double *mag, double *stddev);
    // Returns mag-dev curve
int mfind_getdata(int star, double *jd, double *mag, double *err);
    // Returns light curve of a selected star
int mfind_write(const char *outfile);
    // Computes the results, write output to the file
int mfind_wdata(int star, const char *outfile);
    // Computes the results, write output to the file
int mfind_clean();
    // Frees allocated memory

/* MUNILIST module */

int mlist_init(const char *optfile, int aper, int outfrm, const char *params);
    // Initialize parameteres and open output file
int mlist_read(const char *infile, int frame);
    // Read data from file
int mlist_getnrows(void);
    // Returns number of rows
int mlist_getncols(void);
    // Returns number of columns
int mlist_getcoli(int col, char *buf);
    // Gets column information
int mlist_getrowi(int row, int *frm, double *jd, double *hcor, double *amass);
    // Gets row information
int mlist_getval(int row, int col, double *val, double *err, int *valid);
    // Returns magnitude and its error from specified row, column and aperture
int mlist_delete(int frame);
	// Delete a frame from the list
int mlist_write(const char *outfile);
	// Write data to output file
int mlist_clean(void);
    // Close output file and free allocated memory

/* MUNIPHOT module */

int mphot_init(const char *optfile, const char *apfile);
    // Initializes parameters and read parameter file, aperture file, command file
int mphot(const char *scifile, const char *outfile, double *jd, int *nostar);
    // Reads input file, run protometry, write the output file
int mphot_clean(void);
    // Frees allocated memory

/* TIMEBAT module */

int tcor_init(double tcor);
    // Initializes the parameters
int tcor(const char *sci, const char *out, char *date, char *time);
    // Reads input file, corrects the time and writes output file
int tcor_clean(void);
    // Free allocated memory

/* KOMBINE module */

int kombine_init(const char *optfile);
	// Initializes variables
int kombine_open(const char *outfile);
	// Opens output file
int kombine(const char *filename);
	// Reads the source frame and adds it to the accumulation buffer
int kombine_close(void);
	// Computes the output data and writes it to the file
int kombine_clean(void);
	// Cleans allocated memory

#ifdef __cplusplus
}
#endif

#endif
