/**************************************************************

helcor.c (C-Munipack project)
Copyright (C) 2003 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef _WIN32
#include <io.h>
#endif

#include "comfun.h"
#include "convert.h"
#include "hcor.h"
#include "hterm.h"

void PrintHelp()
{
 printf("HELCOR version %s - utility for the heliocentric correction.\n", VERSION);
 printf(PROJECTLINE);
 printf("\nUsage:helcor [options] ra=hhmmss de=+ddmmss [mask=?] [file1...] \n\n");
 printf("parameters:\n");
 printf("     @dirfile=  file with filenames for conversion\n");
 printf("                (for only '@', read file names from standart input)\n");
 printf("     ra=        right ascension (R.A.) of object \n");
 printf("     dec=       declination (DEC.) of object \n");
 printf("     mask=      output mask (default = %s)\n",DNAME);
 printf("     file1...   names of files to conversion\n");
 printf("                (if no file specified, it runs in terminal mode)\n");
 printf("options:\n");
 printf("     -r         reverse mode (jdhel --> jdgeo)\n");
 printf("     -o<mask>   output mask (equivalent with 'mask=')\n");
 printf("     -c<number> counter initial value (default = 1)\n");
 printf("     -q         quiet mode\n");
 printf("     -h         give this help\n");
 printf("     -v         prints software version\n");
 printf("     -l         display software license\n\n");
 printf(WARRANTYLINE);
 printf(HOMEPAGELINE);
 printf(BUGREPORTS);
}

void report(int status, const char *file)
{
 if (cmpack_debug) 
 	fprintf(stderr, "The following error(s) has been occurred: %d\n",status);
 PrintError(PROGNAME, file, 0, status);
}

void error(int status, const char *file)
{
 report(status, file);
 exit(status);
}

int main(int argc, char **argv)
{
	FILE *f;
	int i, icit, res;
	int reverse = 0, icit0 = 1, isset = 0, level = 0, term = 1, stdinput = 0;
	char *mask = DNAME, *ra = NULL, *de = NULL, *namsci = "";
	char buf1[MAXLINE], line[MAXLINE], format[MAXLINE], outf[MAXLINE], *par;

	/* Zpracovani parametru prikazove radky */
	for (i=1;i<argc;i++) {
	   par = argv[i];
	   /* Prints help */
	   if ((strcmp(par,"-h")==0)||(strcmp(par,"--help")==0)) {
		  PrintHelp();
          return 0;
       }
	   /* Prints version */
	   if ((strcmp(par,"-v")==0)||(strcmp(par,"--version")==0)) {
		  PrintVersion(PROGNAME);
          return 0;
       }
       /* Prints license */
       if ((strcmp(par,"-l")==0)||(strcmp(par,"--license")==0)) {
          PrintLicense();
          return 0;
       }
       /* Flag for input from stdin or dirfile */
       if (strchr(par,'@')==par) {
	       namsci = malloc(strlen(par));
	       strcpy(namsci,&par[1]);
           stdinput = strlen(namsci)==0;
       }
       /* Object's right ascension */
       if (strstr(par,"ra=")==par) {
	       ra = malloc(strlen(par)-2);
	       strcpy(ra,&par[3]);
       }
       /* Object's declination */
       if (strstr(par,"dec=")==par) {
	       de = malloc(strlen(par)-3);
	       strcpy(de,&par[4]);
       }
       /* Mask of output files */
       if (strstr(par,"mask=")==par) {
	       mask = malloc(strlen(par)-4);
	       strcpy(mask,&par[5]);
       }
       if (strstr(par,"-o")==par) {
	       mask = malloc(strlen(par)-1);
	       strcpy(mask,&par[2]);
       }
       /* Initial counter value */
       if (strstr(par,"-c")==par) {
	       icit0 = atol(&par[2]);
       }
       /* Reverse flag */
       if (strstr(par,"-r")==par) {
	       reverse = 1;
       }
       /* Quiet/verbose flags */
       if (strcmp(par,"-q")==0 || strcmp(par,"--quiet")==0 || strcmp(par,"--silent")==0) {
           level = MPK_LEVEL_QUIET;
       }
       if (strcmp(par,"--verbose")==0 || strcmp(par,"--debug")==0) {
           level = MPK_LEVEL_DEBUG;
       }
       /* Conversion mode */
       if ((strchr(par,'=')==0)&&(strchr(par,'@')!=par)&&(strchr(par,'-')!=par)) {
	       term = 0;
       }
  	}

  	cmpack_setoutput(PrintMessage, level);
  	PrintPrologue(PROGNAME);
  
	if (!term) {
  
		// Initialization
  		res = helcor_inits(ra,de,reverse);
  		if (res!=0) error(res, NULL);

		// prepare format string for generating output file names 
		if (strcmp(mask,".")==0) {
	  		mask = NULL;            // dot "." as output mask 
		} else {
	  		MaskToFormat(format,mask);
		}

    	// counter initialization 
    	icit = icit0;
    	isset = 0;

    	// image names reads from command line 
    	for (i=1;i<argc;i++) {
	  		par = argv[i];
      		if ((strchr(par,'=')==0)&&(strchr(par,'@')!=par)&&(strchr(par,'-')!=par)) {

#ifdef _WIN32
				intptr_t fh;
				struct _finddata_t fd;
				if ((fh=_findfirst(par, &fd))!=-1) {
					do {
						if ((fd.attrib & _A_SUBDIR)==0)  {
							if (mask) {
								sprintf(outf,format,icit);
								res=helcor(fd.name,outf);
		          			} else {
		          				res=helcor(fd.name,NULL);
		          			}
		          			if (res!=0) report(res, fd.name);
		          			icit++;
		          		}
					} while (_findnext(fh, &fd)==0);
					_findclose(fh);
		      	 } else {
			      	 report(MPK_CANNOT_OPEN_SRC, par);
		      	 }
#else
				if (mask) {
					sprintf(outf,format,icit);
					res=helcor(par,outf);
				} else {
		        	res=helcor(par,NULL);
				}
				if (res!=0) report(res, par);
				icit++;
#endif
		      }
		}

		// image names reads from dirfile 
	  	if (strlen(namsci)>0) {
		  	f = fopen(namsci,"r");
		  	if (f==NULL) error(MPK_NO_DIRFILE, namsci);
	      	while (fgets(buf1,MAXLINE,f)!=NULL) {
	          	sscanf(buf1,"%s",line);
	          	if (strcmp(line,"")) {
		      	  	if (mask) {
			      	  	sprintf(outf,format,icit);
		      	  	  	res=helcor(line,outf);
	              	} else {
		              	res=helcor(line,NULL);
	              	}
	              	if (res!=0) report(res, line);
	              	icit++;
	          	}
	      	}
	  	}
	
	  	// Image names reads from stdin
	  	if (stdinput) {
	      	while(fgets(buf1,MAXLINE,stdin)!=NULL) {
	          	sscanf(buf1,"%s",line);
	          	if (strcmp(line,"")) {
		      	  	if (mask) {
			      	  	sprintf(outf,format,icit);
		      	  	  	res=helcor(line,outf);
	              	} else {
		              	res=helcor(line,NULL);
	              	}
	              	if (res!=0) report(res, line);
	              	icit++;
	          	}
	      	}
	  	}

	    // Clean up
	    res = helcor_clean();
	    if (res!=0) error(res, NULL);
	    
		if (icit==icit0) 
			PrintOut(0, "No input files.");

	} else {
		
  		// Terminal mode
    	hcor_term(reverse, ra, de);
  }
  
  return 0;
}
