/**************************************************************

konve.c (C-Munipack project)
Copyright (C) 2003 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <fitsio.h>

#include "comfun.h"
#include "fitsfun.h"
#include "inifile.h"
#include "trajd.h"
#include "konve.h"
#include "konvfits.h"
#include "konvsbig.h"
#include "konvoes.h"

/* Registration of known image formats */
typedef struct {
	// Format identifier (FORMAT_xxx constants)
	int format;
	// Function pointers
	konv_testproc *test;
	konv_openproc *open;
	konv_closeproc *close;
	konv_getsproc *gets;
	konv_getlproc *getl;
	konv_getdproc *getd;
	konv_gkynproc *gkyn;
	konv_g2ddproc *g2dd;
	konv_convhdrproc *convhdr;
	konv_convimgproc *convimg;
} fmtstruct;

fmtstruct flist[] = {
	{ MPK_FORMAT_FITS, fits_test, fits_open, fits_close, fits_gets, fits_getl, fits_getd, fits_gkyn, fits_g2dd,
		NULL, NULL },
	{ MPK_FORMAT_SBIG, st_test, st_open, st_close, st_gets, st_getl, st_getd, st_gkyn, st_g2dd,
		st_convhdr, st_convimg },
	{ MPK_FORMAT_OES, oes_test, oes_open, oes_close, oes_gets, oes_getl, oes_getd, oes_gkyn, oes_g2dd,
		oes_convhdr, oes_convimg },
	{ MPK_FORMAT_UNKNOWN }
};

/* File descriptor */
typedef struct {
	void *handle;
	fmtstruct *fmt;
} filestruct;

int  konv_flipx = 0;     /* Flip image in X */
int  konv_flipy = 0;     /* Flip image in Y */
double konv_tcor = 0.0;      /* Time correction in seconds */
char konv_parname[MAXLINE] = ""; /* Parameter file */
int  konv_mode = 0;

int konv_init(const char *parname, int flipx, int flipy, double tcor, int mode)
/* Initializes global variables and reads parameter file */
{
	char buf[MAXLINE];

	/* Init parameters */
	konv_flipx  = flipx!=0;
	konv_flipy  = flipy!=0;
	konv_tcor   = tcor;
	konv_mode   = mode;
	if (parname)
		strncpy(konv_parname,parname,MAXLINE);
	else
		strcpy(konv_parname, "");
    /* Test parameter file */
    if ((mode!=2)&&(mode!=3)&&(strlen(konv_parname)>0)) {
    	FILE *f = fopen(konv_parname,"r");
    	if (f==NULL) return MPK_CANNOT_OPEN_PAR;
    	fclose(f);
		if (strlen(konv_parname)>0) {
		    sprintf(buf,"Parameter file: %s",konv_parname);
		    PrintOut(0, buf);
	    }
    }
  	if (konv_tcor) {
		sprintf(buf,"Time correction: %.3lf seconds.",konv_tcor);
		PrintOut(0, buf);
	}
	if (konv_flipx || konv_flipy) {
		sprintf(buf,"Image flipping: %s%s",(konv_flipx?"X":""),(konv_flipy?"Y":""));
		PrintOut(0, buf);
	}
	return 0;
}

int konv_clean(void)
/* Does nothing - for compatibility only */
{
	return 0;
}

int konv_format(const char *in, int *format)
/* Determines file format. Return 0 if successful, or error code if not. */
{
  FILE *f;
  fmtstruct *fmt;

  *format = MPK_FORMAT_UNKNOWN;
  
  /* Test if source file is present and accessible */
  if (!in) 
 	  return MPK_CANNOT_OPEN_SRC;
  if (strlen(in)==0)
 	  return MPK_CANNOT_OPEN_SRC;
 	  
  f = fopen(in,"rb");
  if (f==NULL)
      return MPK_CANNOT_OPEN_SRC;
  
  /* Test content of file, find image format */
  fmt = flist;
  while (fmt->format != MPK_FORMAT_UNKNOWN) {
  	if (fmt->test) {
		if (fmt->test(f)) {
			*format = fmt->format;
			break;
		}
	}
	fmt++;
  }
  
  /* Close file and return */
  fclose(f);
  return (*format==MPK_FORMAT_UNKNOWN ? MPK_UNKNOWN_FORMAT : 0);
}

/* Returns pointer to format stucture */
fmtstruct *findstruct(int format)
{
  fmtstruct *fmt = flist;
  while (fmt->format != MPK_FORMAT_UNKNOWN) {
  	if (fmt->format == format)
		return fmt;
	fmt++;
  }
  return NULL;
}

int konv_openf(void **file, const char *filename, int format)
{
  /* Create a new file descriptor */
  filestruct *fs = (filestruct*) malloc(sizeof(filestruct));
  if (fs==NULL) { *file = NULL; return MPK_MEMORY; }
  memset(fs, 0, sizeof(filestruct));
  fs->fmt = findstruct(format);
  *file = fs;
  /* Open a source file */
  return fs->fmt->open(&fs->handle, filename);
}

int konv_open(void **file, const char *filename)
{
  int stat, format;
  /* Determine file type */
  stat = konv_format(filename, &format);
  if (stat!=0) { *file = NULL; return stat; }
  /* Open specified file */
  return konv_openf(file, filename, format);
}

int konv_close(void *file)
{
	if (file==0) return MPK_CLOSED_FILE;
	return ((filestruct*)file)->fmt->close(((filestruct*)file)->handle);
}

int konv_gets(void *file, const char *key, char *buf, int buflen)
{
	if (file==0) return MPK_CLOSED_FILE;
	return ((filestruct*)file)->fmt->gets(((filestruct*)file)->handle, key, buf, buflen);
}

int konv_getl(void *file, const char *key, long *val)
{
	if (file==0) return MPK_CLOSED_FILE;
	return ((filestruct*)file)->fmt->getl(((filestruct*)file)->handle, key, val);
}

int konv_getd(void *file, const char *key, double *val)
{
	if (file==0) return MPK_CLOSED_FILE;
	return ((filestruct*)file)->fmt->getd(((filestruct*)file)->handle, key, val);
}

int konv_gkyn(void *file, int index, char *key, int keylen, char *val, int vallen, char *com, int comlen)
{
	if (file==0) return MPK_CLOSED_FILE;
	return ((filestruct*)file)->fmt->gkyn(((filestruct*)file)->handle, index, key, keylen, val, vallen, com, comlen);
}

int konv_g2dd(void *file, double *buf, int flipx, int flipy)
{
	if (file==0) return MPK_CLOSED_FILE;
	return ((filestruct*)file)->fmt->g2dd(((filestruct*)file)->handle, buf, flipx, flipy);
}

/* Print short info */
void printpar(const char *magic, int width, int height, int planes, int bitcount, const char *object, const char *datestr, const char *timestr, double temp, double exp, const char *filter)
{
	char buf[MAXLINE];

	if (strlen(magic)>0) {
		sprintf(buf,"File format: %s", magic);
		PrintOut(0, buf);
	}
	if (width>0 && height>0) {
  		sprintf(buf,"Image size : %d x %d pixels", width, height);
		PrintOut(0, buf);
	}
	if (bitcount>0 && planes>0) {
  		sprintf(buf,"Bit depth  : %d x %d bits", planes, bitcount);
		PrintOut(0, buf);
  	}
	if (strlen(object)>0) {
      	sprintf(buf,"Object     : %s", object);
		PrintOut(0, buf);
	}
	if (strlen(datestr)>0 && strlen(timestr)>0) {
  		sprintf(buf,"Date & time: %s %s UT" , datestr, timestr);
		PrintOut(0, buf);
  	}
	if (exp>=0.0) {
	  	sprintf(buf,"Exposure   : %.2f s", exp);
		PrintOut(0, buf);
  	}
	if (temp>=-999 && temp<=999) {
	  	sprintf(buf,"Temperature: %.1f deg.C", temp);
		PrintOut(0, buf);
  	}
	if (strlen(filter)>0) {
      	sprintf(buf,"Filter     : %s", filter);
		PrintOut(0, buf);
  	}
}

/* Print short info from file structure */
void fileprintpar(filestruct *file)
{
  	char magic[FLEN_VALUE], object[FLEN_VALUE], datestr[FLEN_VALUE], timestr[FLEN_VALUE], filter[FLEN_VALUE];
  	long width, height, planes, bitpix;
  	double temp, exp;

  	if (file->fmt->gets(file->handle, "_magic_", magic, FLEN_VALUE)!=0)
		strcpy(magic, "");
  	if (file->fmt->getl(file->handle, "_width_", &width)!=0)
		width = 0;
	if (file->fmt->getl(file->handle, "_height_", &height)!=0)
		height = 0;
  	if (file->fmt->getl(file->handle, "_planes_", &planes)!=0)
		planes = 1;
	if (file->fmt->getl(file->handle, "_bitpix_", &bitpix)!=0)
		bitpix = 0;
	if (file->fmt->gets(file->handle, "_object_", object, FLEN_VALUE)!=0)
		strcpy(object, "");
  	if (file->fmt->gets(file->handle, "_date_", datestr, FLEN_VALUE)!=0)
		strcpy(datestr, "");
  	if (file->fmt->gets(file->handle, "_time_", timestr, FLEN_VALUE)!=0)
		strcpy(timestr, "");
  	if (file->fmt->getd(file->handle, "_exposure_", &exp)!=0)
		exp = 0.0;
  	if (file->fmt->getd(file->handle, "_temperature_", &temp)!=0)
		temp = -999.9;
  	if (file->fmt->gets(file->handle, "_filter_", filter, FLEN_VALUE)!=0)
		strcpy(filter, "");

	printpar(magic, width, height, planes, bitpix, object, datestr, timestr, temp, exp, filter);
}

void konv_fitsprintpar(fitsfile *fits)
/* Print short info */
{
  	char object[FLEN_VALUE], datestr[FLEN_VALUE], timestr[FLEN_VALUE], filter[FLEN_VALUE];
  	int bitpix, nx, ny;
  	double temp, exp;

	fits_gpar(fits, &nx, &ny, &bitpix);
	fits_getdatetime(fits, datestr, timestr);
	fits_gexp(fits, &exp);
	fits_gtemp(fits, &temp);
	fits_gfilter(fits, filter);
	fits_gobject(fits, object);
	printpar("FITS (Flexible Image Transport System)", nx, ny, 1, bitpix,
		object, datestr, timestr, temp, exp, filter);
}

void includefile(fitsfile *fits, const char *filename, int *stat)
/* Additional header from file */
{
  inifile *par;
  char *key, *val, *com;

  par=iniopen(filename,OPTFILE);
  if (par!=NULL) {
      while (iniread(par,&key,&val,&com)) {
          ffukys(fits,key,val,com,stat);
      }
	  iniclos(par);
  } else {
      *stat = MPK_CANNOT_OPEN_PAR;
  }
}

int FitsPrintHeader(fitsfile *fits)
/* Prints out all parameters in FITS file */
{
  char buf[MAXLINE], key[MAXLINE], val[MAXLINE], com[MAXLINE];
  int i = 1, stat = 0;

  while (ffgkyn(fits,i,key,val,com,&stat)==0) {
      sprintf(buf,"%-20.20s %-24.24s %-32.32s",key,val,com);
	  PrintOut(0, buf);
	  i++;
  }
  return 0;
}

int konv(const char *in, const char *out)
/* Converts 'in' to 'out' */
{
  void *handle;
  fitsfile *dfits, *sfits;
  int  i, sstat=0, dstat=0, format=0;
  char buf[MAXLINE], key[FLEN_KEYWORD], val[FLEN_VALUE], com[FLEN_COMMENT];
  double jd;
  filestruct *file;

  /* Initialization */
  PrintOut(0, "--------------------------------------------");
  sprintf(buf,"File name  : %s",in);
  PrintOut(0, buf);

  /* Determine format of source file */
  sstat = konv_format(in, &format);
  if (sstat!=0) return sstat;

  if (konv_mode==0) {
     /* Convert header and data from source file */
     sprintf(buf,"Output file: %s",out);
     PrintOut(0, buf);
     remove(out);

   	 if (format==MPK_FORMAT_FITS) {
	   	/* Copy source file */
	   	copyfile(in, out);
        /* Open FITS file */
		ffdkopn(&dfits, out, READWRITE, &dstat);
        if (dstat!=0) return dstat;
        /* Flip image */
        if (konv_flipx || konv_flipy) {
		    fits_flipimage(dfits, konv_flipx, konv_flipy, &dstat);
        	if (dstat!=0) return dstat;
		}
		/* Print short info */
		konv_fitsprintpar(dfits);
     } else {
		/* Open source file */
		sstat = konv_openf(&handle, in, format);
		if (sstat!=0) return sstat;
		file = handle;
		/* Print short info */
		fileprintpar(file);
		/* Open destination file */
        ffdkinit(&dfits,out,&dstat);
        if (dstat!=0) return dstat;
		/* Convert header */
		sstat = file->fmt->convhdr(file->handle, dfits);
		if (sstat!=0) return sstat;
		/* Convert image data */
		sstat = file->fmt->convimg(file->handle, dfits, konv_flipx, konv_flipy);
		if (sstat!=0) return sstat;
		/* Close source file */
		konv_close(file);
     }
	 /* Time correction */
	 if (konv_tcor!=0) {
		fits_getjd(dfits, &jd);
		if (jd>0) fits_setjd(dfits, jd + (konv_tcor/86400.0));
	 }
     /* Include parameter file */
     if (strlen(konv_parname)>0)
       	includefile(dfits, konv_parname, &dstat);
     /* Close output and return */
     ffclos(dfits, &dstat);
     return dstat;
 }

 	if (konv_mode==1) {
		/* Converts header only */
		sprintf(buf,"Output file: %s",out);
		PrintOut(0, buf);
		remove(out);
		/* Open destination file */
		ffdkinit(&dfits,out,&dstat);
		if (dstat!=0) return dstat;
		if (format==MPK_FORMAT_FITS) {
			ffdkopn(&sfits, in, READONLY, &sstat);
			if (sstat!=0) {
				ffclos(sfits, &sstat);
				ffclos(dfits, &dstat);
				return sstat;
			}
			/* Copy header only */
			ffcphd(sfits, dfits, &sstat);
			/* Print short info */
			konv_fitsprintpar(dfits);
			ffclos(sfits, &sstat);
		} else {
			/* Open source file */
			sstat = konv_openf(&handle, in, format);
			if (sstat!=0) return sstat;
			file = handle;
			/* Print short info */
			fileprintpar(file);
			/* Open destination file */
			ffdkinit(&dfits,out,&dstat);
			if (dstat!=0) return dstat;
			/* Convert header */
			sstat = file->fmt->convhdr(file->handle, dfits);
			if (sstat!=0) return sstat;
			/* Close source file */
			konv_close(handle);
		}
		/* Time correction */
		if (konv_tcor!=0) {
			fits_getjd(dfits, &jd);
			if (jd>0) fits_setjd(dfits, jd + (konv_tcor/86400.0));
		}
		/* Include parameter file */
		if (strlen(konv_parname)>0)
			includefile(dfits, konv_parname, &dstat);
		/* Close output and return */
		ffclos(dfits, &dstat);
		return dstat;
 	}

	if (konv_mode==2) {
	 	/* Print short info */
	 	sstat = konv_openf(&handle, in, format);
	 	if (sstat!=0) return sstat;
 		fileprintpar(handle);
	 	konv_close(handle);
	 	return 0;
  	}

	if (konv_mode==3) {
  		/* Prints all informations in header */
		sstat = konv_openf(&handle, in, format);
		if (sstat!=0) return sstat;
		file = handle;
      	sprintf(buf,"%-20s %-24s %s","---Key---","---Value---","---Comments---");
      	PrintOut(0, buf);
		i = 1;
		while (file->fmt->gkyn(file->handle, i, key, FLEN_KEYWORD, val, FLEN_VALUE, com, FLEN_COMMENT)==0) {
			sprintf(buf,"%-20.20s %-24.24s %-32.32s",key,val,com);
			PrintOut(0, buf);
			i++;
		}
		konv_close(handle);
		return 0;
	}

 	/* Unknown mode flag */
 	return MPK_INVALID_PAR;
}
