/**************************************************************

fotometr.c (C-Munipack project)
Copyright (C) 2003 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <math.h>

#include "comfun.h"
#include "robmean.h"
#include "phot.h"
#include "phfun.h"
#include "quicksort.h"

#define D(x,y)    d[((x)-1)+(((y)-1)*ncol)]

// Interpolation routine used for computing of FWHM value */
static double inter(double y1, double y, double y2)
{
	if (fabs(y2-y1)<1e-6) return 0.5;
	                 else return (y-y1) / (y2-y1);
}

int PhotSB(void)
// Aperture photometry
{
	double *d, *sky, skymn, skysig, *fwhm, fwhm_med, fwhm_err;
	double apmxsq, rinsq, routsq, maglim, magsq, sumwt, skyvar, sigsq, halfmax;
	double apmag[MAXAP], magerr[MAXAP], area[MAXAP], xc, yc, datum, fractn;
	double skyskw, err1, err2, err3, edge, dysq, rsq;
	double dmag, r, wt;
	double hwhm_a[4], *hwhm_p[4];
	int   apok[MAXAP], skyerr_reported = 0, overexp_reported = 0, badpxl_reported = 0;
	int   ncol, nrow, i, j, k, nmag, lx, ly, mx, my, istar, nsky, ok, nfwhm, hwhm_n;
	StarType *star;

	/* SECTION 1 - Read parameter file, open input and output */

    /* Compute the outermost edge of the largest aperture */
	apmxsq = -1.0;
	for (k=0;k<ph_naper;k++) 
		apmxsq = fmax(apmxsq, (ph_aper[k]+0.5)*(ph_aper[k]+0.5));

	/* Now define the other variables whose values are in the table. */
    rinsq  = ph_is*ph_is;                    /* Inner sky radius squared */
    routsq = ph_os*ph_os;                    /* Outer sky radius squared */
    if (ph_is<=1) {
	    PrintOut(0, "Inner sky radius must be greater than 1.");
	    return MPK_INVALID_PAR;
    }
    if (ph_os<=1) {
	    PrintOut(0, "Outer sky radius must be greater than 1.");
	    return MPK_INVALID_PAR;
    }
    if (routsq > (1.0*MAXSKY/PI + rinsq)) {
	    PrintOut(0, "You have specified too big a sky annulus.");
	    return MPK_INVALID_PAR;
    }
    if (routsq <= rinsq) {
	    PrintOut(0, "Your outer sky radius is no bigger than the inner radius.");
	    return MPK_INVALID_PAR;
    }
	sky  = (double*) malloc(MAXSKY*sizeof(double));
	if (!sky)
		return MPK_MEMORY;
	fwhm = (double*) malloc(ph_nstar*sizeof(double));
	if (!fwhm)
		return MPK_MEMORY;
		
	d    = ph_image.ccd;
	ncol = ph_image.ncol;
	nrow = ph_image.nrow;

	/* Reset input file */
	FirstStar();

	/* SECTION 2 - Derive aperture photometry object by object. */
    maglim= 0.0;
    magsq = 0.0;
    sumwt = 0.0;
    nmag  = 0;
    nfwhm = 0;

    for (istar=0; NextStar(&star); istar++) {
   	    /* Get the coordinates of next object to be measured. */
	    xc   = star->xcen;
	    yc   = star->ycen;
	    /* Initalize magnitudes and errors to default values */
        for (k=0; k<ph_naper; k++) {
	        star->apmag[k] = 99.999;
	        star->magerr[k] = 9.999;
        }
        /* Compute the limits of the submatrix. */
        lx   = (int) fmax(1, (int)(xc-ph_os+0.5)+1);
        mx   = (int) fmin(ncol, (int)(xc+ph_os+0.5));
        ly   = (int) fmax(1, (int)(yc-ph_os+0.5)+1);
        my   = (int) fmin(nrow, (int)(yc+ph_os+0.5));
        edge = fmin(xc-0.5, fmin((ncol+0.5)-xc, fmin(yc-0.5, (nrow+0.5)-yc)));
	    /* Initialize star counts and aperture area. */
        for (k=0;k<ph_naper;k++) {
            apmag[k]  = 0.0;
            magerr[k] = 0.0;
            area[k]   = 0.0;
	        apok[k]   = (ph_aper[k]>0)&&(edge>=ph_aper[k]);
        }
        /* Now read through the submatrix, picking out the data we want. */
        nsky = 0;
        for (j=ly;j<=my;j++) {
            dysq = (j - yc)*(j - yc);
            for (i=lx;i<=mx;i++) {
                rsq = dysq + (i - xc)*(i - xc);    /* squared distance from [i,j] to the center of the star */
                datum = D(i,j);
                if (rsq>=rinsq && rsq<=routsq && nsky<MAXSKY && datum>ph_minvalue && datum<ph_maxvalue) {
	                // This pixel is in the sky annulus
                    sky[nsky] = datum;
                    nsky++;
                }
                if (rsq<=apmxsq) {
	                /* this pixel fits to largest aperture radius */
                    r = sqrt(rsq) - 0.5;          /* distance from the center */
                    for (k=0;k<ph_naper;k++) {
                        if (r <= ph_aper[k]) {
                            fractn = fmax(0.0, fmin(1.0, ph_aper[k]-r));
                            if (datum >= ph_maxvalue) {
	                            /* bad datum value */
	                            if (!overexp_reported) {
		    						PrintOut(0, "There are overexposed pixels in the aperture.");
		                            overexp_reported = 1;
	                            }
	                            apok[k] = 0;
                            }
                            if (datum <= ph_minvalue) {
	                            /* bad datum value */
	                            if (!badpxl_reported && datum<=0) {
		    						PrintOut(0, "There are bad pixels in the aperture.");
		                            badpxl_reported = 1;
	                            }
	                            apok[k] = 0;
                            }
                            apmag[k] += fractn*datum;   /* sum of pixel values that fits to the aperture k */
                            area[k]  += fractn;         /* size of the area of the aperture k */
                        }
                    }
                }
            }
        }
	    if (nsky<MINSKY) {
		    if (!skyerr_reported) {
		    	PrintOut(0, "There aren't enough pixels in the sky annulus. Are you sure your bad pixel thresholds are all right? If so, then you need a larger outer sky radius.");
		    	skyerr_reported = 1;
	    	}
		    continue;
	    }
	    /*  Obtain the mode, standard deviation, and skewness of the peak in the sky histogram */
	    cmpack_robustmeanD(nsky, sky, &skymn, &skysig);
	    skyskw = 0.0;
        skyvar=skysig*skysig;    /* variance of the sky brightness */
        sigsq =skyvar/nsky;      /* square of the standard error of the mean sky brightness */
        star->skymod = skymn;
        star->skysig = skysig;
        star->skyskw = skyskw;
        /* Subtract the sky from the integrated brightnesses in the apertures, convert the results */
        /* to magnitudes, and compute standard errors. Detect bad values. */
        for (k=0;k<ph_naper;k++) {
	        ok = 0;
            if (apok[k] && skysig>=-0.5) { 
	            apmag[k] = apmag[k] - skymn*area[k];
            	if (apmag[k]>0.0) { 
            		err1 = area[k]*skyvar;
            		err2 = apmag[k]/ph_image.phpadu;
            		err3 = sigsq*area[k]*area[k];
            		magerr[k] = fmin(9.999, 1.0857*sqrt(err1 + err2 + err3)/apmag[k]);
            		apmag[k] = 25.0 - 2.5*log10(apmag[k]);
            		ok = (apmag[k]>-99 && apmag[k]<99);
        		}
    		}
    		if (ok) {
       			star->apmag[k] = apmag[k];
        		star->magerr[k] = magerr[k];
    		} else {
       			star->apmag[k] = 99.999;
        		star->magerr[k] = 9.999;
    		}	
        }
        if ((star->apmag[0]<99.0)&&(star->magerr[0]<9.0)) {
	        /* Compute FWHM of the star */
	        halfmax = (star->height + star->skymod) / 2.0;
	        hwhm_n = 0;
	        for (i=star->xmax-1, j=star->ymax; i>0; i--) {
		        if (D(i,j)<halfmax && D(i+1,j)>=halfmax) {
			        hwhm_a[hwhm_n++] = star->xcen - i - inter(D(i,j), halfmax, D(i+1,j));
			        break;
		        }
	        }
	        for (i=star->xmax+1, j=star->ymax; i<=ncol; i++) {
		        if (D(i,j)<halfmax && D(i-1,j)>=halfmax) {
			        hwhm_a[hwhm_n++] = i - star->xcen - inter(D(i,j), halfmax, D(i-1,j));
			        break;
		        }
	        }
	        for (i=star->xmax, j=star->ymax-1; j>0; j--) {
		        if (D(i,j)<halfmax && D(i,j+1)>=halfmax) {
			        hwhm_a[hwhm_n++] = star->ycen - j - inter(D(i,j), halfmax, D(i,j+1));
			        break;
		        }
	        }
	        for (i=star->xmax, j=star->ymax+1; j<=nrow; j++) {
		        if (D(i,j)<halfmax && D(i,j-1)>=halfmax) {
			        hwhm_a[hwhm_n++] = j - star->ycen - inter(D(i,j), halfmax, D(i,j-1));
			        break;
		        }
	        }
	        if (hwhm_n<2) {
		        star->fwhm = -1;
	        } else {
		        /* Median FWHM */
		        for (i=0; i<hwhm_n; i++) 
		        	hwhm_p[i] = hwhm_a + i;
				QuickSortD(hwhm_a, hwhm_n, (void**) hwhm_p);
				if (hwhm_n & 0x1) star->fwhm = 2.0 * hwhm_a[hwhm_n/2];
				             else star->fwhm = hwhm_a[hwhm_n/2-1] + hwhm_a[hwhm_n/2];
        	    if (star->fwhm>0) {
	        	    fwhm[nfwhm++] = star->fwhm;
        	    }
		    }
	        /* Compute magnidute limit */
	        if (star->magerr[0]>0.00001) {
	        	dmag    = -2.5*log10(star->height/ph_image.thres);
            	wt      = (2.0/(2.0 - dmag))*(100.0/star->magerr[0])*(100.0/star->magerr[0]);
            	maglim += wt*(star->apmag[0] - dmag);
            	magsq  += wt*(star->apmag[0] - dmag)*(star->apmag[0] - dmag);
            	sumwt  += wt;
            	nmag++;
        	}
    	}
    }

    /* Compute magnitude limit */
    if (sumwt>0.0) {
        maglim = maglim/sumwt;
        magsq  = magsq/sumwt - maglim*maglim;
        magsq  = sqrt(fmax(0.0, magsq));
        ph_image.maglim = maglim;
        ph_image.magsq  = magsq;
    } else {
	    ph_image.maglim = 99.999;
	    ph_image.magsq  = 9.999;
    }
    /* Compute robust mean FWHM value */
    if (nfwhm>0) {
	    cmpack_robustmeanD(nfwhm, fwhm, &fwhm_med, &fwhm_err);
	    ph_image.fwhm_med = fwhm_med;
	    ph_image.fwhm_err = fwhm_err;
    } else {
	    ph_image.fwhm_med = 0.0;
	    ph_image.fwhm_err = 0.0;
    }

    /* Normal return */
    free(sky);
    free(fwhm);
    return 0;
}
